/* --------------------------------------------------------------- */
/* lpcI2CIO.c -- I2C device functions for LPC                      */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* Low level functions to effect I2C bus I/O.                      */
/*                                                                 */
/* The channels are by default assigned speeds in order to allow   */
/* the use of both slow and faster devices:                        */
/*                                                                 */
/* * channel0: 400 kbit/s                                          */
/* * channel1: 100 kbit/s                                          */
/* * channel2: 400 kbit/s                                          */
/*                                                                 */
/* These speeds can be changed using i2cInitIO before i2cIO is     */
/* first called for a channel.                                     */
/* --------------------------------------------------------------- */
// Not tested for channel 0 [not accessible on mbed]

#define TOLLOS_IO             // allow use of I/O functions
#include "tollosI2C.h"
#include "lpc17xx_i2c.h"
#include "lpc17xx_pinsel.h"


// constants
#define I2C_CHANNELS 3        // Total number of channels that might
                              // be present
#define I2C_CHANNEL0 1        // 1 if I2C0 accessible, 0 if not
#define I2C_CHANNEL1 1        // 1 if I2C1 accessible
#define I2C_CHANNEL2 1        // 1 if I2C2 accessible

#define I2C_SPEED0 400000     // default I2C0 speed
#define I2C_SPEED1 100000     // default I2C1 speed
#define I2C_SPEED2 400000     // default I2C2 speed

// statics
static int  businit[I2C_CHANNELS];                // 1 once initialized
static uint busspeed[I2C_CHANNELS]=               // speed of each bus
            {I2C_SPEED0, I2C_SPEED1, I2C_SPEED2}; // [init to defaults]
static LPC_I2C_TypeDef * const bus[I2C_CHANNELS]= // LPC addresses
            {LPC_I2C0, LPC_I2C1, LPC_I2C2};       //   ..

/* --------------------------------------------------------------- */
/* i2cInitIO -- initialize an I2C channel (hardware interface)     */
/*                                                                 */
/*   channel -- channel number to initialize                       */
/*   hz      -- speed to run the channel                           */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* This function is called automatically, with a default speed     */
/* setting, when i2cIO is called and the channel has not been      */
/* initialized.  Therefore it need only be invoked explicitly for  */
/* a non-default speed.                                            */
/* --------------------------------------------------------------- */
int i2cInitIO(uint channel, uint hz) {
  PINSEL_CFG_Type pincfg;               // for configuring pins

  // default: I2C interfaces enabled
  //          I2C clock sources CCLK/4

  // setup the pins according to channel
  pincfg.Portnum=PINSEL_PORT_0;              // I2C pins are all port 0
  pincfg.OpenDrain=PINSEL_PINMODE_OPENDRAIN; // as per $19.1
  pincfg.Pinmode=PINSEL_PINMODE_TRISTATE;    // (external pullup)

  switch(channel) {

    #if I2C_CHANNEL0
    case 0: {
      pincfg.Funcnum=PINSEL_FUNC_1;     // function 1 = I2C0
      pincfg.Pinnum=PIN27;              // SDA0
      PINSEL_ConfigPin(&pincfg);
      pincfg.Pinnum=PIN28;              // SCL0
      PINSEL_ConfigPin(&pincfg);
      break;}
    #endif

    #if I2C_CHANNEL1
    case 1: {
      pincfg.Funcnum=PINSEL_FUNC_3;     // function 3 = I2C1
      pincfg.Pinnum=PIN0;               // SDA1
      PINSEL_ConfigPin(&pincfg);
      pincfg.Pinnum=PIN1;               // SCL1
      PINSEL_ConfigPin(&pincfg);

      // SDA1 & SCL1 are also available on p0.19/F3 and p0.20/F4
      break;}
    #endif

    #if I2C_CHANNEL2
    case 2: {
      pincfg.Funcnum=PINSEL_FUNC_2;     // function 2 = I2C2
      pincfg.Pinnum=PIN10;              // SDA2
      PINSEL_ConfigPin(&pincfg);
      pincfg.Pinnum=PIN11;              // SCL2
      PINSEL_ConfigPin(&pincfg);
      break;}
    #endif

    default: return -1;                 // bad channel
    } // switch

  I2C_Init(bus[channel], hz);           // bus initialize, with speed
  I2C_Cmd(bus[channel], ENABLE);        // bus activate

  busspeed[channel]=hz;                 // save the speed requested
  businit[channel]=1;                   // now initialized
  return 0;
  } // i2cInitIO

/* --------------------------------------------------------------- */
/* i2cIO -- I2C general I/O function                               */
/*                                                                 */
/*   device -- channel and physical address of device              */
/*             (channel shifted left 16 bits and ORed with 7-bit   */
/*             or 10-bit I2C physical address); currently only     */
/*             7-bit addresses are supported                       */
/*   put    -- start of byte(s) to write                           */
/*   putlen -- count of byte(s) to write [may be 0]                */
/*   get    -- start of byte(s) to read                            */
/*   getlen -- count of byte(s) to read [may be 0]                 */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* put is ignored if putlen is 0 (NULL recommended).               */
/* get is ignored if getlen is 0 (NULL recommended).               */
/*                                                                 */
/* This function will automatically initialize the appropriate     */
/* I2C channel when first called for a channel, unless already     */
/* initialized by an explicit call to i2cInitIO.                   */
/* --------------------------------------------------------------- */
int i2cIO(uint device, byte *put, uint putlen, byte *get, uint getlen) {
  uint channel=i2cChannel(device); // channel 0-2
  uint addr=i2cAddr7Bit(device);   // physical address
  I2C_M_SETUP_Type blok;           // Master setup control block
  Status mtd;                      // value from LPC functions

  if (channel>=I2C_CHANNELS) return -1; // minimal checking

  // initialise channel if needed
  if (!businit[channel]) {
   if (i2cInitIO(channel, busspeed[channel])) return -2;
   }

  // fill in control block
  blok.sl_addr7bit=addr;
  blok.tx_data=put;
  blok.tx_length=putlen;
  blok.rx_data=get;
  blok.rx_length=getlen;
  blok.retransmissions_max=5;

  mtd=I2C_MasterTransferData(bus[channel], &blok, I2C_TRANSFER_POLLING);
  if (mtd==ERROR) return -1;
  return 0;  // SUCCESS
  } // i2cIO
