/* --------------------------------------------------------------- */
/* lpcSPI9IO.c -- write-only 9 bit SPI glue                        */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* This module wraps the interface to the LPC SPI routines using   */
/* some fixed assumptions:                                         */
/*                                                                 */
/*   -- write only, Master mode                                    */
/*   -- 12 MHz (SPI peripheral clock = 96 MHz), or otherwise as    */
/*      fast as permitted (1/8 system clock)                       */
/*   -- MSB first                                                  */
/*   -- 9-bit data                                                 */
/*   -- clock polarity high, sample on rising edge                 */
/*   -- three-wire interface; SSEL is the chip select line/port    */
/*                                                                 */
/* This might be generalized later but currently this is intended  */
/* to be used by other glue routines, not directly from Tollos.    */
/*                                                                 */
/* Polling output is used here because the overhead of using       */
/* interrupts would not give a useful power or speed saving (at    */
/* 12 MHz each word being written is only 9 clocks, or 750ns).     */
/*                                                                 */
/* --------------------------------------------------------------- */


#define  TOLLOS_IO                 // allow use of I/O functions
#include "tollos.h"
#include "LPC17xx.h"
#include "lpc17xx_spi.h"
#include "lpc17xx_clkpwr.h"
#include "lpc17xx_pinsel.h"
#include "lpc17xx_gpio.h"


#define SPICLOCK 12000000          // SPI clock frequency

// port and bit to use for Chip Select (SSEL)
#define CS_PORT PORT0
#define CS_MASK BIT16
#define CS_PIN  PIN16


// local function
static int spi9Send(flag cd, byte *data, uint length);

/* --------------------------------------------------------------- */
/* spi9CommandIO -- send a one-byte Command                        */
/*                                                                 */
/*   b -- byte to send                                             */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* --------------------------------------------------------------- */
int spi9CommandIO(byte b) {
  return spi9Send(0, &b, 1);
  } // spi9CommandIO

/* --------------------------------------------------------------- */
/* spi9Command2IO -- send a two-byte Command                       */
/*                                                                 */
/*   b1 -- first byte to send                                      */
/*   b2 -- second byte to send                                     */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* --------------------------------------------------------------- */
int spi9Command2IO(byte b1, byte b2) {
  byte buff[2];
  buff[0]=b1;
  buff[1]=b2;
  return spi9Send(0, buff, 2);
  } // spi9Command2IO

/* --------------------------------------------------------------- */
/* spi9DataIO -- send byte(s) as Data                              */
/*                                                                 */
/*   data   -- address of data to send                             */
/*   length -- count of bytes to send                              */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* --------------------------------------------------------------- */
int spi9DataIO(byte *data, uint length) {
  return spi9Send(1, data, length);
  } // spi9DataIO

/* --------------------------------------------------------------- */
/* spi9InitIO -- initialize SPI and third wire                     */
/*                                                                 */
/* SSEL is the GPIO line that will be used for the chip select     */
/* third wire of the 3-wire 9 bit interface; this could be         */
/* parameterized later to allow multiple devices to be driven on   */
/* the SPI bus, providing they all use the same protocol.          */
/*                                                                 */
/* See the module header for other assumptions.                    */
/*                                                                 */
/* --------------------------------------------------------------- */
void spi9InitIO(void) {
  PINSEL_CFG_Type pincfg;               // for configuring pins
  SPI_CFG_Type spicfg;                  // ditto SPI


  // setup the SPI pins; there is only one channel
  pincfg.Portnum=PINSEL_PORT_0;         // SPI pins are all port 0
  pincfg.OpenDrain=PINSEL_PINMODE_NORMAL;
  pincfg.Pinmode=PINSEL_PINMODE_TRISTATE;
  pincfg.Funcnum=PINSEL_FUNC_3;         // function 3 = SPI

  // SPI pins (set up MISO correctly, although not used here)
  pincfg.Pinnum=PIN15;                  // SCK
  PINSEL_ConfigPin(&pincfg);

  pincfg.Pinnum=PIN17;                  // MISO
  PINSEL_ConfigPin(&pincfg);

  pincfg.Pinnum=PIN18;                  // MOSI (SDA)
  PINSEL_ConfigPin(&pincfg);

  // set up SSL (as GPIO) CS (Chip Select) pin
  pincfg.Portnum=CS_PORT;               // will be 0
  pincfg.Funcnum=PINSEL_FUNC_0;         // function 0 = GPIO
  pincfg.Pinnum=CS_PIN;                 // SSEL
  PINSEL_ConfigPin(&pincfg);
  GPIO_SetDir(CS_PORT, CS_MASK, 1);     // set as output

      GPIO_ClearValue(CS_PORT, CS_MASK);

  // set up SPI device

  spicfg.Databit=SPI_DATABIT_9;         // 9 bits: C/D then 8
  spicfg.CPHA=SPI_CPHA_SECOND;          // phase sample on second edge
  spicfg.CPOL=SPI_CPOL_LO;              // polarity active low, sample on rise
  spicfg.Mode=SPI_MASTER_MODE;          // master, write only
  spicfg.DataOrder=SPI_DATA_MSB_FIRST;  // most significant first

  // clock rate is SPICLOCK if possble, or system clock/8 if lower
  spicfg.ClockRate=min(SPICLOCK, SystemCoreClock/8);

  // set the peripheral clock for SPI to be system clock /1
  // (the default is system clock /4)
  CLKPWR_SetPCLKDiv(CLKPWR_PCLKSEL_SPI, CLKPWR_PCLKSEL_CCLK_DIV_1);

  SPI_Init(LPC_SPI, &spicfg);           // initalize

  } // spi9InitIO

/* --------------------------------------------------------------- */
/* spi9Send -- send byte(s) as 9-bit data or commands              */
/*                                                                 */
/*   cd     -- command/data flag: Command=0, Data=1                */
/*   data   -- address of data to send                             */
/*   length -- count of bytes to send  [0=nop]                     */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* This does not use the LPC SPI_ReadWrite function because the    */
/* latter expects 16-bit slots for 9-bit data, which would         */
/* require copying all the data to twice the size.  Local code     */
/* also permits toggling the CS line between bytes.                */
/*                                                                 */
/* CS toggling between every 9 bit word significantly slows down   */
/* data transfer however (x 2 for a 6MHz SPI clock, and x 3 for    */
/* 12 MHz), so is not effected here.  Instead the caller may       */
/* choose to send large amounts of data in bursts.                 */
/*                                                                 */
/* With the code here and an SPI clock of 12 MHz the actual data   */
/* transfer rate measured during send is one 9-bit byte every      */
/* 1.25 us (800kB/s).                                              */
/* --------------------------------------------------------------- */
static int spi9Send(flag cd, byte *data, uint length) {
  uint work;                       // work word
  uint commask=cd<<8;              // shifted command/data bit
  byte *bp;                        // byte pointer
  byte *bend=data+length;          // byte pointer endpoint
  int  rc=0;                       // return code
  uint status;                     // status word

  GPIO_ClearValue(CS_PORT, CS_MASK);    // pull down CS line

  work=LPC_SPI->SPSR;                   // read to clear status

  for (bp=data; bp<bend; bp++) {             // each byte
    LPC_SPI->SPDR=commask | *bp;             // send 9-bit data

    for (;;) {                               // wait for Tx complete
      status=LPC_SPI->SPSR;
      if (status & SPI_SPSR_SPIF) break;
      }

    // check for errors
    work=status & (SPI_SPSR_ABRT|SPI_SPSR_MODF|SPI_SPSR_ROVR|SPI_SPSR_WCOL);
    if (work) {
      rc=-((int)work);   // negate the error bits for post mortem
      if (work & SPI_SPSR_MODF) {            // mode fault ...
        work=LPC_SPI->SPCR;                  // requires write to
        LPC_SPI->SPCR=work;                  // .. SPCR to clear
        }
      break;                                 // abandon send
      } // error
    } // byte loop

  work=LPC_SPI->SPDR;                        // read to clear final SPIF

  GPIO_SetValue(CS_PORT, CS_MASK);           // finished with device

  return rc;
  } // spi9Send

/* --------------------------------------------------------------- */
/* spi9StopIO -- stop the SPI device                               */
/*                                                                 */
/* This just disables the SPI clock.                               */
/* --------------------------------------------------------------- */
void spi9StopIO(void) {
  SPI_DeInit(LPC_SPI);
  } // spi9StopIO

