/* --------------------------------------------------------------- */
/* tollosGyro.c -- gyroscope functions for Tollos                  */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These functions provide basic gyroscope functions for           */
/* Tollos, intended to be device-independent (although at the      */
/* moment an I2C connection is assumed).                           */
/*                                                                 */
/* To use, the device must be initialized using gyroInit.  This    */
/* indicates which I2C channel the device is on and which port/pin */
/* is used for the DataReady interrupt.  The gyroscope range is    */
/* determined by the device used.  (For example, for the ITG3200   */
/* it is +/- 2000 degrees/sec.)                                    */
/*                                                                 */
/* The latest gyroscope reading can be read safely using           */
/* gyroGetXYZ, which returns the three axes in milli-degrees/sec.  */
/* The axis directions are as defined by the device manufacturer.  */
/* Note that gyroscopes will generally have some constant offset,  */
/* or drift, that will indicate a rate of turn on a 'static'       */
/* device; this may be partly compensated for by adjusting for the */
/* die temperature, available using gyroTemperature.               */
/*                                                                 */
/* The default update rate is selected by the device driver code   */
/* and will be at least 25Hz (that is, the data will be no more    */
/* than 40ms old).  Update is not necessarily synchronous with the */
/* Tollos tick rate.   The rate can be changed with gyroRate; the  */
/* low pass-filter bandwidth is set to match or exceed the rate.   */
/*                                                                 */
/* tollosGyro will buffer up to four gyroscope readings.           */
/* This allows an update rate of double the application's          */
/* wake-up frequency, plus one missed wakeup.  An application      */
/* should therefore call gyroGetXYZ repeatedly until 'no data      */
/* available' is reported in order to clear queued data and to     */
/* avoid overrun.                                                  */
/*                                                                 */
/* --------------------------------------------------------------- */
// Temperature compensation is a 'big win'; experimental code for
// it is running and probably will be moved here in due course.

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

// vector of readings from Gyroscope
typedef struct {
  int x;                      // X reading [md/s]
  int y;                      // Y reading [md/s]
  int z;                      // Z reading [md/s]
  } gyroData;

// Ring-buffer of readings; this is used conventionally to avoid race
// conditions between the interrupt handler and the gyroGetXYZ
// function.
//
// In particular:
//
//   * The write pointer (index) is only incremented (and wrapped if
//     need be) altered by the interrupt handler, after good data
//     have been read.
//
//   * The read pointer is only incremented (and wrapped if need be)
//     after the Get function has copied data to the caller.
//
//   * If the read and write pointers are equal, the ring is empty.
//
//   * If stepping the write pointer would make the write pointer
//     equal the read pointer then an overrun has occurred and the
//     write pointer is not stepped.  This is reported (and cleared)
//     at the next read.
//
// Hence: the two levels of code always write to and read from
// different entries in the ring, and each level independently
// modifies the pointer that it 'owns'.
//
// The default size of the buffer ring allows for four entries.
#define RINGSIZE 4

static gyroData gyroRing[RINGSIZE+1];   // +1 for the working slot
static int      gyroWrite=0;            // write index
static int      gyroRead=0;             // read index

// Status flags
static flag     gyroReady=0;            // gyroInit has been successful
static flag     gyroOverrun=0;          // overrun has occurred

// local function
static void gyroHandler(void);     // interrupt handler


/* --------------------------------------------------------------- */
/* gyroGetXYZ -- read latest x, y, and z angular rates             */
/*                                                                 */
/*   x, y, z -- receive angular rates [milli-degrees/sec:          */
/*              1000 = 1.000 degrees/second]; or are unchanged if  */
/*              no data are available.                             */
/*                                                                 */
/*   returns 0 if data good, +1 if data is good but an overrun has */
/*     occurred since the last read, -1 if no good data from read, */
/*     or -2 if the device is not initialized.                     */
/*                                                                 */
/* Results are returned scaled to milli-degrees/second.            */
/*                                                                 */
/* When +1 (overrun) is returned the overrun status is cleared and */
/* future good reads will return 0 until another overrun occurs.   */
/*                                                                 */
/* This must not be called from an interrupt level with higher     */
/* priority than the gyroscope interrupt.                          */
/* --------------------------------------------------------------- */
int gyroGetXYZ(int *x, int *y, int*z) {
  // see above for the procedure used here

  if (!gyroReady) return -2;            // unready
  if (gyroWrite==gyroRead) return -1;   // ring empty

  // return data from the 'read' slot
  *x=gyroRing[gyroRead].x;
  *y=gyroRing[gyroRead].y;
  *z=gyroRing[gyroRead].z;

  gyroRead++;                           // bump index
  if (gyroRead==RINGSIZE+1) gyroRead=0; // wrap
  if (!gyroOverrun) return 0;           // is good
  gyroOverrun=0;                        // had overrun; clear flag
  return 1;                             // .. and report
  } // gyroGetXYZ

/* --------------------------------------------------------------- */
/* gyroHandler -- interrupt handler, called when Data Ready        */
/*                                                                 */
/* This reads the data into the 'write' slot, then when and if     */
/* bumps the write index unless this would overrun.                */
/* --------------------------------------------------------------- */
static void gyroHandler(void) {
  int rc;                          // work
  int newWrite;                    // tentative write index

  rc=gyroGetXYZIO(&gyroRing[gyroWrite].x, &gyroRing[gyroWrite].y,
                 &gyroRing[gyroWrite].z);
  // interrupt is cleared by the read
  if (rc<0) return;                          // bad data

  newWrite=gyroWrite+1;                      // bump index
  if (newWrite==RINGSIZE+1) newWrite=0;      // wrap

  if (newWrite!=gyroRead) gyroWrite=newWrite;// OK to use it
   else gyroOverrun=1;                       // [don't bump]
  } // gyroHandler

/* --------------------------------------------------------------- */
/* gyroInit -- configure and initialize gyroscope                  */
/*                                                                 */
/*   channel -- I2C channel on which device is to be found         */
/*   drport  -- port for the Data Ready interrupt pin              */
/*   drpin   -- pin number (not bit mask) for the Data Ready       */
/*              interrupt pin                                      */
/*                                                                 */
/*   returns 0 if initialized OK, negative if error                */
/*                                                                 */
/* This sets up the pin and interrupt handler for the Data Ready   */
/* interrupt from the device then calls the device driver          */
/* initialization function to start the device.                    */
/*                                                                 */
/* The gyroInitIO function will set:                               */
/* * an update rate of at least 25 Hz (but not necessarily higher  */
/* than that, to save power); this may be changed with gyroRate    */
/* * an interrupt on Data Ready (this is handled by tollosGyro).   */
/*                                                                 */
/* Note that it can be 50ms or more before the first valid data    */
/* are available, so a gyroGetXYX call before that time may report */
/* no data available, as might a call that reads the device less   */
/* than 25ms after the last read.                                  */
/* --------------------------------------------------------------- */
int gyroInit(uint channel, uint drport, uint drpin) {
  int rc;                     // work

  gyroWrite=0;                // clear indexes ..
  gyroRead=0;                 // ..

  // set up interrupt pin from number (bitFuns use a mask)
  rc=bitEnable(drport, 1<<drpin, BIT_RISE, gyroHandler);
  if (rc==0) rc=gyroInitIO(channel);
  if (rc==0) gyroReady=1;    // record good status
  return rc;
  } // gyroInit

/* --------------------------------------------------------------- */
/* gyroRate -- change update rate of gyroscope                     */
/*                                                                 */
/*   rate -- rate requested [Hz, 1..200]                           */
/*                                                                 */
/*   returns the rate chosen if the rate was changed OK, or -1     */
/*     if error (rate zero or too high, or a bus error)            */
/*                                                                 */
/* The gyroscope will be set to a slowest available rate that      */
/* is >= the requested rate, to minimize resource usage.  The      */
/* low-pass filter (if any) bandwidth is set to match or exceed    */
/* the chosen rate.                                                */
/*                                                                 */
/* To ensure working with 100kHz I2C, the maximum rate is limited  */
/* to 200Hz -- but note that this may use significant processor    */
/* bandwidth.                                                      */
/* --------------------------------------------------------------- */
int gyroRate(uint rate) {
  return gyroRateIO(rate);    // pass on and back
  } // gyroRate

/* --------------------------------------------------------------- */
/* gyroStop -- stop and power-down gyroscope                       */
/*                                                                 */
/*   returns 0 if gyroscope stopped OK, negative if an error       */
/*     or it was not running.                                      */
/* --------------------------------------------------------------- */
int gyroStop(void) {
  int rc;                     // work
  gyroReady=0;                // invalidate any data
  rc=gyroStopIO();            // stop the device
  gyroRead=0;                 // be clean
  gyroWrite=0;                // ..
  return rc;
  } // gyroStop

/* --------------------------------------------------------------- */
/* gyroTemperature -- return current gyroscope die temperature     */
/*                                                                 */
/*   returns the latest temperature reading in milli-degrees       */
/*     Celsius, or GYRO_NOTEMP if no reading is available          */
/*                                                                 */
/* Many gyroscope devices provide the temperature of the chip die, */
/* which can be used to compensate for drift due to temperature    */
/* change.  GYRO_NOTEMP is returned if no temperature reading is   */
/* available or the Gyro device is stopped.                        */
/* --------------------------------------------------------------- */
int gyroTemperature(void) {
  return gyroTemperatureIO(); // pass back
  } // gyroTemperature

